#include <cstdlib>
#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/time.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <pthread.h>
#include <errno.h>
#include <time.h>
#include <semaphore.h>
#include <map>

#include <vector>
#include <string>
#include <set>
#include <iostream>
#include <fstream>
#include <sstream>

using namespace std;

#define BEGIN 1
#define ADD 2
#define COMMIT 3
#define SLEEP 4

struct Request {
	int idClienta;
	int row;
};

struct Response {
	int row;
	int value;
	int timestamp;
	timeval sendTime;
	timeval receiveTime;
};

struct ChangeRows {
	int row;
	int value;
	int timestamp;
	timeval receiveTime;
};

struct ResultCommit {
	bool commit;
	timeval timeCommit;
};

struct Instruction {
	int type;
	int value1;
	int value2;
	int value3;
};

vector<Instruction> loadedInstructions;

int server_sock;
socklen_t addr_len;
struct sockaddr_in remote_addr, my_addr;
ofstream clientLog;
string fileLogClient;
int idClienta;
int port;

timeval getTime() {
	timeval cas;
	gettimeofday(&cas, NULL);
	return cas;
}

string timevalToString(timeval time) {
	stringstream ss;
	ss << time.tv_sec << "." << time.tv_usec;
	return ss.str();
}

string numberToString(int number) {
	stringstream ss;
	ss << number;
	return ss.str();
}

void writeToLog(timeval receiveData, timeval casPrijeti, string data) {
	clientLog.open(fileLogClient.c_str(), std::ios_base::app);
	clientLog << "- dst: {p: SERVER, ts: " << timevalToString(casPrijeti)
			<< "}\n";
	clientLog << "  src: {p: P" << numberToString(idClienta) << ", ts: "
			<< timevalToString(receiveData) << "}\n";
	clientLog << "  data: '" << data << "'\n";
	clientLog.close();
}

void writeToLogFromServer(timeval receiveData, timeval casPrijeti,
		string data) {
	clientLog.open(fileLogClient.c_str(), std::ios_base::app);
	clientLog << "- src: {p: SERVER, ts: " << timevalToString(receiveData)
			<< "}\n";
	clientLog << "  dst: {p: P" << numberToString(idClienta) << ", ts: "
			<< timevalToString(casPrijeti) << "}\n";
	if (data.compare("COMMIT") == 0) {
		clientLog << "  color: '#00FF00'\n";
		clientLog << "  data: 'COMMIT'\n";
	} else if (data.compare("ABORT") == 0) {
		clientLog << "  color: '#FF0000'\n";
		clientLog << "  data: 'ABORT'\n";
	} else {
		clientLog << "  data: '" << data << "'\n";
	}
	clientLog.close();
}

void loadFile(string fileName) {
	ifstream fin(fileName.c_str());

	char rows[80];

	while (!fin.eof()) {
		fin.getline(rows, 80);

		string instruction = string(rows);
		Instruction ins;
		if (instruction.find("BEGIN") == 0) {
			ins.type = BEGIN;
		} else if (instruction.find("ADD") == 0) {
			ins.type = ADD;
			sscanf(instruction.c_str(), "ADD %d %d %d", &ins.value1,
					&ins.value2, &ins.value3);
		} else if (instruction.find("COMMIT") == 0) {
			ins.type = COMMIT;
		} else if (instruction.find("SLEEP") == 0) {
			ins.type = SLEEP;
			sscanf(instruction.c_str(), "SLEEP %d", &ins.value1);
		}

		if (instruction.compare("") != 0) {
			loadedInstructions.push_back(ins);
		}

	}
	fin.close();
}

void connectToServer() {
	char ip[32] = "127.0.0.1";
	server_sock = socket(AF_INET, SOCK_STREAM, 0);
	if (server_sock <= 0) {
		cout << "Socket ER\n";
		exit(EXIT_FAILURE);
	}
	memset(&my_addr, 0, sizeof(struct sockaddr_in));

	my_addr.sin_family = AF_INET;
	my_addr.sin_port = htons(port);
	my_addr.sin_addr.s_addr = inet_addr(ip);

	int return_value = connect(server_sock, (struct sockaddr *) &my_addr,
			sizeof(struct sockaddr_in));
	if (return_value == 0)
		printf("Connected to server\n");
	else
		printf("Connection failed to server\n");
}

void disconnectFromServer() {
	close(server_sock);
}

void processing() {
	vector<Instruction> processingInstructionBeforeCommit;
	set<int> changesLines;
	for (unsigned int i = 0; i < loadedInstructions.size(); i++) {
		if (loadedInstructions[i].type == SLEEP) {
			cout << "processing SLEEP\n";
			sleep(loadedInstructions[i].value1 / 10);
		}
		if (loadedInstructions[i].type == ADD) {
			cout << "processing ADD\n";
			changesLines.insert(loadedInstructions[i].value1);
			changesLines.insert(loadedInstructions[i].value2);
			changesLines.insert(loadedInstructions[i].value3);

			processingInstructionBeforeCommit.push_back(loadedInstructions[i]);
		}
		if (loadedInstructions[i].type == BEGIN) {
			cout << "processing BEGIN\n";
			connectToServer();
		}
		if (loadedInstructions[i].type == COMMIT) {
			map<int, Response> mapaParametru;

			timeval casOdeslaniBegin;
			casOdeslaniBegin = getTime();
			for (set<int>::iterator i = changesLines.begin();
					i != changesLines.end(); i++) {
				int radek = *i;
				//odeslat parametry
				Request tmp;
				tmp.idClienta = idClienta;
				tmp.row = radek;
				write(server_sock, (void *) &tmp, sizeof(Request));
				changesLines.erase(i);
			}
			Request endReq;
			endReq.idClienta = -1;
			write(server_sock, (void *) &endReq, sizeof(Request));

			/**
			 * ted jsem pozadal radky co chci..
			 */

			Response par;
			//timeval receiveTime;
			timeval receiveData;
			while (true) {
				read(server_sock, (void *) &par, sizeof(Response));

				if (par.row == -1)
					break;

				//receiveTime = getTime();
				receiveData = par.receiveTime;
				mapaParametru[par.row] = par;
			}
			sleep(1);
			writeToLogFromServer(receiveData, getTime(), "DATA");
			//writeToLog(casOdeslaniBegin, timeBegin, "BEGIN");

			//spocitat a odeslat
			timeval timeSendCommit = getTime();
			for (unsigned int j = 0;
					j < processingInstructionBeforeCommit.size(); j++) {
				int cislo1 =
						mapaParametru[processingInstructionBeforeCommit[j].value1].value;
				int cislo2 =
						mapaParametru[processingInstructionBeforeCommit[j].value2].value;

				Response zmena =
						mapaParametru[processingInstructionBeforeCommit[j].value3];

				ChangeRows zh;
				zh.receiveTime = timeSendCommit;
				zh.timestamp = zmena.timestamp;
				zh.row = zmena.row;
				zh.value = cislo1 + cislo2;
				write(server_sock, (void *) &zh, sizeof(ChangeRows));
			}
			ChangeRows zh;
			zh.row = -1;
			write(server_sock, (void *) &zh, sizeof(ChangeRows));

			processingInstructionBeforeCommit.clear();

			//prijmout potvreznei commitu
			ResultCommit vs;
			read(server_sock, (void *) &vs, sizeof(ResultCommit));

			sleep(1);
			writeToLogFromServer(vs.timeCommit, getTime(),
					vs.commit ? "COMMIT" : "ABORT");

			if (vs.commit == true) {
				cout << "COMMIT\n";
			} else {
				cout << "ABORT\n";
			}
			disconnectFromServer();
		}
	}
}

int main(int argc, char* argv[]) {

	if (argc != 4) {
		cout << "./Client <port> <id> <soubor>\n";
		exit(1);
	}

	// proot socket retarder je retarder
	sleep(1);

	port = atoi(argv[1]);
	idClienta = atoi(argv[2]);
	char* fileName = argv[3];

	cout << "The client id - " << idClienta << " started\n";
	srand(time(NULL) + idClienta);
//	int delay = rand()%3;
//	sleep(delay);
//	cout << "Sleep na - " << delay << " ID - " << idClienta << "\n";
	fileLogClient = "../log/log_client" + numberToString(idClienta) + ".txt";
	cout << "The log file has been created\n";
	loadFile(fileName);
	cout << "The file has been loaded - " << loadedInstructions.size()
			<< " instructions in " << idClienta << " client\n";
	sleep(2);
	processing();

	cout << "The end...\n";
	return 0;
}

